#include "ui.h"
#include "gpu.h"
#include "draw.h"

#include "arpa/inet.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"

#include "ft2build.h"
#include FT_FREETYPE_H

VkResult render_thread(GLFWwindow* window, RenderContext* render) {
  FT_Library library;

  UIContextStorage ui;
  UIContainerStorage container;

  FontStorage font;
  uint32_t font_index;
  TextureStorage texture;
  uint32_t texture_index;

  VkBuffer transfer;
  VmaAllocation transfer_memory;
  void* mapped;

  VkResult result;

  VK_RESULT(create_ui_context(
        10,
        10,
        render,
        &ui));

  if(FT_Init_FreeType(&library) != FT_Err_Ok) {
    return VK_ERROR_UNKNOWN;
  }

  VK_RESULT(load_font("test.ttf", 16, VK_TRUE, library, render, &ui, &font_index, &font));

  VK_RESULT(load_texture("test.png", render, &ui,  &texture_index, &texture));

  UILayerInput layers[] = {
    {
      .strings = 0,
      .codes = 0,
      .drawables = 1,
    },
    {
      .strings = 1,
      .codes = 100,
      .drawables = 1,
      .font = font_index,
    },
  };

  VK_RESULT(create_container(0xDEADBEEF, 0.0, 0.0, 200.0, 200.0, sizeof(layers)/sizeof(UILayerInput), layers, render, &container));

  VK_RESULT(create_transfer_buffer(render->allocator, 2*sizeof(UIDrawable) + sizeof(UIString) + 100*sizeof(uint32_t), &transfer, &transfer_memory, &mapped));

  UIDrawable* mapped_drawable = (UIDrawable*)(mapped + 0);
  set_ui_image(0.0, 0.0, 100.0, 200.0, 1.0, 1.0, 1.0, 1.0, texture_index, &mapped_drawable[0]);
  set_ui_rect(100.0, 0.0, 100.0, 200.0, 0.0, 1.0, 0.0, 1.0, &mapped_drawable[1]);

  UIString* mapped_string = (UIString*)(mapped + 2*sizeof(UIDrawable));
  set_ui_string(0.0, 100.0, 32.0, 1.0, 1.0, 1.0, 1.0, 13, 0, &mapped_string[0]);
  uint32_t* mapped_codes = (uint32_t*)(mapped + 2*sizeof(UIDrawable) + sizeof(UIString));
  VK_RESULT(set_ui_codes("Hello, World!", mapped_codes, 0, font.charmap, font.num_symbols));

  VkCommandBuffer command_buffer = command_begin_single(render->device, render->transfer_pool);
  command_copy_buffer(command_buffer, transfer, container.layers[0].drawables, 0, 0, sizeof(UIDrawable));
  command_copy_buffer(command_buffer, transfer, container.layers[1].drawables, sizeof(UIDrawable), 0, sizeof(UIDrawable));

  command_copy_buffer(command_buffer, transfer, container.layers[1].strings, 2*sizeof(UIDrawable), 0, sizeof(UIString));
  command_copy_buffer(command_buffer, transfer, container.layers[1].codes, 2*sizeof(UIDrawable) + sizeof(UIString), 0, 100*sizeof(uint32_t));
  VK_RESULT(command_end_single(render->device, command_buffer, render->transfer_pool, render->transfer_queue));
  vkQueueWaitIdle(render->transfer_queue.handle);
  destroy_transfer_buffer(render->allocator, transfer, transfer_memory);

  while(glfwWindowShouldClose(window) == 0) {
    glfwPollEvents();

    result = draw_frame(render, &ui, container.layers, container.layer_count);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      return result;
    }
  }
  return 0;
}

int logic_thread() {
  return 0;
}

int network_thread() {
  return 0;
}

int main() {
  GLFWwindow* window = init_window();
  if(window == NULL) {
    return 1;
  }

  RenderContext render = {};
  if(init_vulkan(window, &render) != VK_SUCCESS) {
    return 2;
  }

  if(render_thread(window, &render) != VK_SUCCESS) {
    return 3;
  }

  return 0;
}
